#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Iterable

__all__ = ["Topic"]


class Topic(tuple):
    """A Topic describes the location withing the logging hierarchy.

    The top level is the empty topic.

    Separator is ``/``. Leading and trailing whitespace is removed for each component.
    Components are forced to be strings.

    A Topic is considered to contain OTHER Topic, if all elements of topic are in OTHER (starting at the first).
    """

    def __new__(cls, topic: str | Iterable[str] = ""):
        if isinstance(topic, str):
            topic = topic.split("/")
        topic = [str(i).strip() for i in topic]
        return super().__new__(cls, tuple([i for i in topic if i]))

    def __contains__(self, other: "Topic") -> bool:
        if not len(self):
            # The empty Topic is the root and contains every other Topic
            return True
        if len(self) > len(other):
            # Topic has more elements, therefore other CANNOT be in Topic
            return False
        return all([i == j for i, j in zip(self, other)])

    def __str__(self):
        return "/".join(self)

    def __eq__(self, other):
        return str(self) == str(other)



import typing

import click

# noinspection PyProtectedMember
from click.decorators import _param_memo

if typing.TYPE_CHECKING:
    import typing_extensions

    P = typing_extensions.ParamSpec("P")

R = typing.TypeVar("R")
T = typing.TypeVar("T")
_AnyCallable = typing.Callable[..., typing.Any]
FC = typing.TypeVar("FC", bound=typing.Union[_AnyCallable, click.Command])


def make_decorator(cls, param_decls, **attrs):

    def decorator(f: FC) -> FC:
        _param_memo(f, cls(param_decls, **attrs))
        return f

    return decorator



import inspect

from hakisto import logger
from ._logger_globals import logger_globals

logger_globals.register_excluded_source_file(inspect.currentframe().f_code.co_filename)


__all__ = [
    "hakisto_file",
    "hakisto_inline_location",
    "hakisto_process_all",
    "hakisto_process_file",
    "hakisto_process_inline_location",
    "hakisto_process_severity",
    "hakisto_process_short_trace",
    "hakisto_process_short_trace",
    "hakisto_severity",
    "hakisto_short_trace",
    "hakisto_use_color",
]

try:
    # noinspection PyPackageRequirements
    import click  # noqa: F401
except ImportError as e:
    logger.critical("package 'click' not found")
    raise e from None
else:
    from .click import (
        hakisto_file,
        hakisto_inline_location,
        hakisto_process_all,
        hakisto_process_file,
        hakisto_process_inline_location,
        hakisto_process_severity,
        hakisto_process_short_trace,
        hakisto_severity,
        hakisto_short_trace,
        hakisto_use_color,
    )



import threading

from ._logger_globals import logger_globals
from .severity import Severity


class HandlerSettings:
    """Settings for logging to be applied to a specific Handler"""

    def __init__(
        self,
        severity: Severity | int | str = None,
        output_to_file: bool = None,
        use_color: bool = None,
        date_format: str = None,
        short_trace: bool = None,
        color_palette: int = None,
        inline_location: set[Severity] = None,
    ) -> None:
        if severity is not None and not isinstance(severity, Severity):
            severity = Severity(severity)
        self._severity = severity
        self._output_to_file = output_to_file
        self._use_color = use_color
        self._short_trace = short_trace
        self._inline_location = inline_location
        self._date_format = date_format
        self._color_palette = color_palette
        self._lock = threading.Lock()

    @property
    def severity(self) -> Severity:
        if self._severity is None:
            return logger_globals.severity
        return self._severity

    @severity.setter
    def severity(self, value: Severity | int | str):
        if not isinstance(value, Severity):
            value = Severity(value)
        with self._lock:
            self._severity = value

    @property
    def color_palette(self) -> int:
        if self._color_palette is None:
            return logger_globals.color_palette
        return self._color_palette

    @color_palette.setter
    def color_palette(self, value: int) -> None:
        with self._lock:
            self._color_palette = value

    @property
    def date_format(self) -> str:
        if self._date_format is None:
            return logger_globals.date_format
        return self._date_format

    @date_format.setter
    def date_format(self, value: str) -> None:
        with self._lock:
            self._date_format = value

    @property
    def output_to_file(self) -> bool:
        if self._output_to_file is None:
            return logger_globals.output_to_file
        return self._output_to_file

    @output_to_file.setter
    def output_to_file(self, value: bool):
        with self._lock:
            self._output_to_file = value

    @property
    def use_color(self) -> bool:
        if self._use_color is None:
            return logger_globals.use_color
        return self._use_color

    @use_color.setter
    def use_color(self, value: bool):
        with self._lock:
            self._use_color = value

    @property
    def short_trace(self) -> bool:
        if self._short_trace is None:
            return logger_globals.short_trace
        return self._short_trace

    @short_trace.setter
    def short_trace(self, value: bool):
        with self._lock:
            self._short_trace = value

    @property
    def inline_location(self) -> set:
        if self._inline_location is None:
            return logger_globals.inline_location
        return self._inline_location



import os
from typing import Type, Any, Callable

import click

from ._click import FC, make_decorator
from .._logger import Logger


def hakisto_use_color(
    *param_decls: str,
    default: bool = None,
    show_default: bool = True,
    cls: Type[click] | None = None,
    **attrs: Any
) -> Callable[[FC], FC]:
    """Attaches option to use short trace (source list for last only).

    This is a boolean option, with the default provided, taken from HAKISTO_USE_COLOR or ``True``.

    All positional arguments are  passed as parameter declarations to :class:`Option`,
    if none are provided, defaults to ``--log-short-trace/--no-log-short-trace``.

    Keyword arguments not described below are passed as keyword arguments to :func:`click.Option`,
    arguments are forwarded unchanged (except ``cls``).

    :param default:
    :param show_default:
    :param cls: the option class to instantiate.  This defaults to :class:`Option`.
    :param param_decls: Passed as positional arguments to the constructor of ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """

    attrs["default"] = default if default is not None else os.getenv("HAKISTO_USE_COLOR", "TRUE").upper() in (
        "TRUE",
        "ON",
        "1",
        "YES",
    )
    attrs["show_default"] = show_default
    attrs["help"] = attrs.get("help", "Use color?")

    return make_decorator(
        cls=cls or click.Option,
        param_decls=param_decls or ("--log-use-color/--no-log-use-color",),
        **attrs
    )


def hakisto_process_use_color(log_use_color: bool, **kwargs) -> None:
    """Process the Hakisto use color option.

    :param log_use_color:
    """
    Logger.global_use_color = log_use_color



import os
from typing import Sequence, Type, Any, Callable

import click

from ._click import FC, make_decorator
from .._logger_globals import logger_globals
from ..severity import Severity


def hakisto_severity(
    *param_decls: str,
    default: str = None,
    choices: Sequence[str] = None,
    case_sensitive: bool = False,
    show_default: bool = True,
    cls: Type[click.Option] | None = None,
    **attrs: Any,
) -> Callable[[FC], FC]:
    """Attaches HAKISTO SEVERITY option to the command.

    This is always a ``click.Choice`` option, with the default taken from HAKISTO_SEVERITY or ``INFO``.

    The default options are the Hakisto Severities (not case-sensitive).

    All positional arguments are  passed as parameter declarations to :class:`Option`,
    if none are provided, defaults to ``--log, --log-severity``.

    Keyword arguments not described below are passed as keyword arguments to :func:`click.Option`,
    arguments are forwarded unchanged (except ``cls``).

    :param default: default Hakisto Severity.
    :param choices: List of permitted Hakisto Severities.
    :param case_sensitive:
    :param show_default:
    :param cls: the option class to instantiate.  This defaults to :class:`Option`.
    :param param_decls: Passed as positional arguments to the constructor of ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """
    attrs["default"] = (default or os.getenv("HAKISTO_SEVERITY", "INFO")).upper()
    attrs["show_default"] = show_default
    attrs["type"] = click.Choice(
        [i.upper() for i in choices or Severity.names.keys()], case_sensitive=case_sensitive
    )
    attrs["help"] = attrs.get("help", "Minimum Logging Severity")

    return make_decorator(
        cls=cls or click.Option,
        param_decls=param_decls or ("--log", "--log-severity", "log_severity"),
        **attrs,
    )


def hakisto_process_severity(log_severity: str, **kwargs) -> None:
    """Process the Hakisto Severity option.

    :param log_severity:
    """
    logger_globals.severity = log_severity



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import threading


from .severity import Severity


class LoggerGlobals:
    """Singleton to store global information for logging"""

    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(LoggerGlobals, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self._severity = Severity(
            os.getenv("HAKISTO_SEVERITY", "").strip().upper() or 0
        )
        self._output_to_file = os.getenv(
            "HAKISTO_ENABLE_FILE", "True"
        ).strip().upper() in ("TRUE", "ON", "1", "YES")
        self._use_color = os.getenv("HAKISTO_USE_COLOR", "True").strip().upper() in (
            "TRUE",
            "ON",
            "1",
            "YES",
        )
        self._short_trace = os.getenv(
            "HAKISTO_SHORT_TRACE", "False"
        ).strip().upper() in ("TRUE", "ON", "1", "YES")
        hakisto_inline_location = set(
            [
                i.strip()
                for i in os.getenv("HAKISTO_INLINE_LOCATION", "").upper().split()
            ]
        )
        if "ALL" in hakisto_inline_location:
            hakisto_inline_location = set(
                [v for k, v in Severity.values.items() if k > 0]
            )
        self._inline_location = set([Severity(i) for i in hakisto_inline_location])
        self._date_format = os.getenv(
            "HAKISTO_SHORT_TRACE", "%y-%m-%d %H:%M:%S"
        ).strip()
        self._color_palette = {
            "LIGHT": 0,
            "DARK": 1,
        }.get(os.getenv("HAKISTO_COLORS", "").strip().upper(), 0)

        self._excluded_source_files = set()
        self._handlers = set()

    def register_excluded_source_file(self, file_name: str) -> None:
        """This must be called in the source file of any descendent to exclude the respective
        entries in the call-stack.

        Recommendation: Call on source file level (module).

        .. code:: python

           logger_globals.register_excluded_source_file(inspect.currentframe().f_code.co_filename)

        :param file_name: Source file name
        """
        with self._lock:
            self._excluded_source_files.add(file_name)

    @property
    def excluded_source_files(self) -> set[str]:
        """Get a copy of excluded_source_files when identifying *real* caller in call-stack."""
        return self._excluded_source_files.copy()

    @property
    def color_palette(self) -> int:
        return self._color_palette

    @color_palette.setter
    def color_palette(self, value: int) -> None:
        if value < 0 or value > 1:
            value = 0
        with self._lock:
            self._color_palette = value

    @property
    def date_format(self) -> str:
        return self._date_format

    @date_format.setter
    def date_format(self, value: str) -> None:
        with self._lock:
            self._date_format = value

    @property
    def severity(self) -> Severity:
        return self._severity

    @severity.setter
    def severity(self, value: Severity | int | str) -> None:
        if not isinstance(value, Severity):
            value = Severity(value)
        with self._lock:
            self._severity = value

    @property
    def output_to_file(self) -> bool:
        return self._output_to_file

    @output_to_file.setter
    def output_to_file(self, value: bool) -> None:
        with self._lock:
            self._output_to_file = value

    @property
    def use_color(self) -> bool:
        return self._use_color

    @use_color.setter
    def use_color(self, value: bool) -> None:
        with self._lock:
            self._use_color = value

    @property
    def short_trace(self) -> bool:
        return self._short_trace

    @short_trace.setter
    def short_trace(self, value: bool) -> None:
        with self._lock:
            self._short_trace = value

    @property
    def inline_location(self) -> set:
        return self._inline_location

    def add_handler(self, handler) -> None:
        with self._lock:
            self._handlers.add(handler)

    def drop_handler(self, handler) -> None:
        with self._lock:
            self._handlers.discard(handler)

    @property
    def handlers(self) -> set:
        return self._handlers


logger_globals = LoggerGlobals()



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Iterable

from .handler_settings import HandlerSettings
from .topic import Topic


class _LoggerHandlerBase:
    """Common base for Logger and Handler"""

    def __init__(
            self,
            topic: Topic | str | Iterable[str] = "",
            settings: HandlerSettings = None,
            **kwargs,
    ) -> None:
        self._topic = Topic(topic)
        self.settings = settings or HandlerSettings()
        self._is_active = True

    @property
    def is_active(self) -> bool:
        return self._is_active

    def activate(self, active: bool = True) -> None:
        self._is_active = active

    def deactivate(self) -> None:
        self.activate(active=False)



from typing import Callable, Any, Type, Sequence

import click

from ._click import FC, make_decorator
from ..severity import Severity
from .._logger_globals import logger_globals


def hakisto_inline_location(
    *param_decls: str,
    choices: Sequence[str] = None,
    case_sensitive: bool = False,
    cls: Type[click.Option] | None = None,
    **attrs: Any,
) -> Callable[[FC], FC]:
    """Attaches option to include the location in the identifier (will not be clickable), for the respective
    Severity. Can be used multiple times.

    All positional arguments are  passed as parameter declarations to :class:`Option`,
    if none are provided, defaults to ``--log-inline-location``.

    Keyword arguments not described below are passed as keyword arguments to :func:`click.Option`,
    arguments are forwarded unchanged (except ``cls``).

    :param choices: List of permitted Hakisto Severities.
    :param case_sensitive:
    :param cls: the option class to instantiate.  This defaults to :class:`Option`.
    :param param_decls: Passed as positional arguments to the constructor of ``cls``.
    :param attrs: Passed as keyword arguments to the constructor of ``cls``.
    """

    attrs["multiple"] = True
    if not choices:
        choices = list(Severity.names.keys())
        choices.append("ALL")
    choices = [i.upper() for i in choices]
    attrs["type"] = click.Choice(choices, case_sensitive=case_sensitive)
    attrs["help"] = attrs.get("help", "Show file location inline for severity")

    return make_decorator(
        cls=cls or click.Option,
        param_decls=param_decls or ("--log-inline-location",),
        **attrs,
    )


def hakisto_process_inline_location(log_inline_location: tuple, **kwargs) -> None:
    """Process the Hakisto show location option."""

    if "ALL" in log_inline_location:
        for k in Severity.values:
            if k:
                logger_globals.inline_location.add(Severity(k))

    else:
        for i in log_inline_location:
            logger_globals.inline_location.add(Severity(i))



from .file import hakisto_file, hakisto_process_file
from .inline_location import hakisto_inline_location, hakisto_process_inline_location
from .severity import hakisto_severity, hakisto_process_severity
from .short_trace import hakisto_short_trace, hakisto_process_short_trace
from .use_color import hakisto_use_color, hakisto_process_use_color

__all__ = [
    "hakisto_file",
    "hakisto_inline_location",
    "hakisto_process_all",
    "hakisto_process_file",
    "hakisto_process_inline_location",
    "hakisto_process_severity",
    "hakisto_process_short_trace",
    "hakisto_severity",
    "hakisto_short_trace",
    "hakisto_use_color",
]

PROCESSOR = dict(
    log_file=hakisto_process_file,
    log_inline_location=hakisto_process_inline_location,
    log_severity=hakisto_process_severity,
    log_short_trace=hakisto_process_short_trace,
    log_use_color=hakisto_process_use_color,
)


def hakisto_process_all(**kwargs):
    """Process all Hakisto options."""

    if "log_file" in kwargs:
        hakisto_process_file(kwargs["log_file"])
    if "log_inline_location" in kwargs:
        hakisto_process_inline_location(kwargs["log_inline_location"])
    if "log_severity" in kwargs:
        hakisto_process_severity(kwargs["log_severity"])
    if "log_short_trace" in kwargs:
        hakisto_process_short_trace(kwargs["log_short_trace"])
    if "log_use_color" in kwargs:
        hakisto_use_color(kwargs["log_use_color"])


__import__("pkg_resources").declare_namespace(__name__)  # pragma: no cover



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#  Handlers (listeners) are attached using pypubsub, but we want to isolate this
#  from other usages.

from types import MappingProxyType

from .colors import (
    LIGHT_MAGENTA,
    LIGHT_CYAN,
    LIGHT_GREEN,
    LIGHT_YELLOW,
    LIGHT_RED,
    MAGENTA,
    CYAN,
    GREEN,
    YELLOW,
    RED,
)

__all__ = [
    "TRACE",
    "DEBUG",
    "VERBOSE",
    "INFO",
    "SUCCESS",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "Severity",
]

_T = 100
_D = 200
_V = 300
_I = 400
_S = 500
_W = 600
_E = 700
_C = 800


class Severity(int):
    values = MappingProxyType(
        {
            0: "NOTSET",
            _T: "TRACE",
            _D: "DEBUG",
            _V: "VERBOSE",
            _I: "INFO",
            _S: "SUCCESS",
            _W: "WARNING",
            _E: "ERROR",
            _C: "CRITICAL",
        }
    )
    names = MappingProxyType({v: k for k, v in values.items()})
    colors = MappingProxyType(
        {
            _T: (LIGHT_MAGENTA, MAGENTA),
            _D: (LIGHT_CYAN, CYAN),
            _V: (LIGHT_GREEN, GREEN),
            _I: (LIGHT_GREEN, GREEN),
            _S: (LIGHT_GREEN, GREEN),
            _W: (LIGHT_YELLOW, YELLOW),
            _E: (LIGHT_RED, RED),
            _C: (LIGHT_RED, RED),
        }
    )

    def __new__(cls, value):
        if not isinstance(value, int):
            if value.upper() not in cls.names:
                raise ValueError(f"{value} is not a valid severity name")
            value = cls.names[value.upper()]
        return super().__new__(cls, value)

    def __str__(self):
        return self.values.get(self, super().__str__())

    def color(self, palette: int = 0) -> str:
        """Get respective color from palette"""
        i = min(self // 100, 8) * 100
        if not i:
            return ""
        return self.colors[i][palette]


# fmt: off
TRACE    = Severity(_T)  # More information than regular DEBUG (e.g. variable values).
DEBUG    = Severity(_D)  # Debugging (technical) information.
VERBOSE  = Severity(_V)  # Additional (non-technical) information.
INFO     = Severity(_I)  # Information.
SUCCESS  = Severity(_S)  # Successful execution.
WARNING  = Severity(_W)  # Warning that might require attention.
ERROR    = Severity(_E)  # Significant issue. Processing might continue.
CRITICAL = Severity(_C)  # Critical issue. Usually processing is aborted.
# fmt: on



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import inspect
import sys
from types import FrameType, TracebackType


from ._base import _LoggerHandlerBase
from ._logger_globals import logger_globals

from .severity import (
    TRACE,
    DEBUG,
    VERBOSE,
    INFO,
    SUCCESS,
    WARNING,
    ERROR,
    CRITICAL,
    Severity,
)
from .subject import Subject


__all__ = ["Logger"]


class Logger(_LoggerHandlerBase):
    """Main Logger class. Can be inherited from, but should be sufficient for most situations.

    """

    def critical(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log a **CRITICAL** entry.

        ``CRITICAL`` entries will include the respective source location, source section and local variables.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            CRITICAL,
            message_id=message_id,
            force_location=force_location,
            *args,
            **kwargs,
        )

    def error(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log an **ERROR** entry.

        ``ERROR`` entries will include the respective source location.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            ERROR, message_id=message_id, force_location=force_location, *args, **kwargs
        )

    def warning(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log a **WARNING** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            WARNING,
            message_id=message_id,
            force_location=force_location,
            *args,
            **kwargs,
        )

    def success(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log a **SUCCESS** entry.

        This has been added to support responses from SAP.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            SUCCESS,
            message_id=message_id,
            force_location=force_location,
            *args,
            **kwargs,
        )

    def info(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log an **INFO** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            INFO, message_id=message_id, force_location=force_location, *args, **kwargs
        )

    def verbose(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log a **VERBOSE** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            VERBOSE,
            message_id=message_id,
            force_location=force_location,
            *args,
            **kwargs,
        )

    def debug(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log a **DEBUG** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            DEBUG, message_id=message_id, force_location=force_location, *args, **kwargs
        )

    def trace(
        self, *args: str, message_id: str = None, force_location: bool = False, **kwargs
    ) -> None:
        """Log a **TRACE** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        self.log(
            TRACE, message_id=message_id, force_location=force_location, *args, **kwargs
        )

    def log(
        self,
        severity,
        *args: str,
        message_id: str = None,
        force_location: bool = False,
        **kwargs,
    ) -> None:
        """Log an entry.

        While this method has been made public, be advised, that using integers directly might break in the
        future, if the implementation is modified.

        :param severity: The severity level of the log entry.
        :param args: The Message(s). Every message will create a separate entry.
        :param message_id: Message ID.
        :param force_location: Force output of location
        """
        if not all((self._is_active, self.settings.severity <= severity)):
            return

        frame = self._get_caller()
        for message in args:
            subject = Subject(
                topic=self._topic,
                severity=severity,
                frame=frame,
                message=str(message),
                message_id=message_id,
                force_location=force_location,
                **kwargs,
            )
            for handler in logger_globals.handlers:
                # noinspection PyBroadException
                try:
                    handler(subject)
                except:
                    pass

    @staticmethod
    def _get_caller() -> FrameType:
        """Return *real* caller.

        If this method is overridden, make sure that the right frames are excluded.

        :meta public:
        """
        frame = inspect.currentframe()
        while frame.f_code.co_filename in logger_globals.excluded_source_files:
            frame = frame.f_back
        return frame

    def __repr__(self):
        return f"Logger({self._topic})"


def log_exception(exception_class, exception: Exception, trace_back: TracebackType):
    """Hook to handle uncaught exceptions.

    The entry is sent to **all** Handlers.

    A :class:`imuthes.logging.Handler` **must** implement ``handle_exception`` when it should react to this.

    :param exception_class: Not used
    :type exception_class: object
    :param exception:
    :param trace_back:
    """
    subject = Subject(
        severity=Severity(sys.maxsize), frame=trace_back, message=str(exception)
    )
    for handler in logger_globals.handlers:
        # noinspection PyBroadException
        try:
            handler.handle_exception(subject)
        except:
            pass


logger_globals.register_excluded_source_file(inspect.currentframe().f_code.co_filename)
sys.excepthook = log_exception
